from meross_iot.cloud.abilities import *
from meross_iot.cloud.device import AbstractMerossDevice
from meross_iot.logger import POWER_PLUGS_LOGGER as l
from meross_iot.meross_event import DeviceDoorStatusEvent
from threading import Event

class GenericGarageDoorOpener(AbstractMerossDevice):
    # Channels
    _channels = []

    # Dictionary {channel_id (door) -> status}
    _door_state = None

    def __init__(self, cloud_client, device_uuid, **kwords):
        super(GenericGarageDoorOpener, self).__init__(cloud_client, device_uuid, **kwords)

    def get_status(self):
        with self._state_lock:
            if self._door_state is None:
                self._get_status_impl()

        return self._door_state

    def _handle_push_notification(self, namespace, payload, from_myself=False):
        def fire_garage_door_state_change(dev, channel_id, o_state, n_state, f_myself):
            if o_state != n_state:
                evt = DeviceDoorStatusEvent(dev=dev, channel_id=channel_id, door_state=n_state,
                                            generated_by_myself=f_myself)
                self.fire_event(evt)

        with self._state_lock:
            if namespace == GARAGE_DOOR_STATE:
                for door in payload['state']:
                    channel_index = door['channel']
                    state = door['open'] == 1
                    old_state = self._door_state[channel_index]
                    self._door_state[channel_index] = state
                    fire_garage_door_state_change(self, channel_index, old_state, state, from_myself)

            elif namespace == REPORT:
                # For now, we simply ignore push notification of these kind.
                # In the future, we might think of handling such notification by caching them
                # and avoid the network round-trip when asking for power consumption (if the latest report is
                # recent enough)
                pass

            else:
                l.error("Unknown/Unsupported namespace/command: %s" % namespace)

    def _get_status_impl(self):
        if self._door_state is None:
            self._door_state = {}
        data = self.get_sys_data()['all']
        if 'digest' in data:
            for c in data['digest']['garageDoor']:
                self._door_state[c['channel']] = c['open'] == 1
        return self._door_state

    def _get_channel_id(self, channel):
        # Otherwise, if the passed channel looks like the channel spec, lookup its array indexindex
        if channel in self._channels:
            return self._channels.index(channel)

        # if a channel name is given, lookup the channel id from the name
        if isinstance(channel, str):
            for i, c in enumerate(self.get_channels()):
                if c['devName'] == channel:
                    return c['channel']

        # If an integer is given assume that is the channel ID
        elif isinstance(channel, int):
            return channel

        # In other cases return an error
        raise Exception("Invalid channel specified.")

    def _operate_door(self, channel, state, callback, wait_for_sensor_confirmation):
        # If the door is already in the target status, do not execute the command.
        already_in_state = False
        with self._state_lock:
            already_in_state = self.get_status()[channel] == state

        if already_in_state and callback is None:
            l.info("Command was not executed: the door state is already %s" % ("open" if state else "closed"))
            return
        elif already_in_state and callback is not None:
            callback(None, self._door_state[channel])
            return

        payload = {"state": {"channel": channel, "open": state, "uuid": self.uuid}}
        if wait_for_sensor_confirmation:
            door_event = None
            if callback is None:
                door_event = Event()

            def waiter(data):
                self.unregister_event_callback(waiter)
                if data.channel != channel:
                    return
                if callback is None:
                    door_event.set()
                else:
                    if data.door_state != state:
                        callback("Operation failed", data.door_state)
                    else:
                        callback(None, data.door_sate)

            self.register_event_callback(waiter)
            self.execute_command(command="SET", namespace=GARAGE_DOOR_STATE, payload=payload, callback=None)

            if callback is None:
                door_event.wait()
                current_state = self._door_state[channel]
                if current_state != state:
                    raise Exception("Operation failed.")

        else:
            self.execute_command(command="SET", namespace=GARAGE_DOOR_STATE, payload=payload, callback=callback)

    def open_door(self, channel=0, callback=None, ensure_opened=True):
        c = self._get_channel_id(channel)
        return self._operate_door(c, 1, callback=callback, wait_for_sensor_confirmation=ensure_opened)

    def close_door(self, channel=0, callback=None, ensure_closed=True):
        c = self._get_channel_id(channel)
        return self._operate_door(c, 0, callback=callback, wait_for_sensor_confirmation=ensure_closed)

    def get_channels(self):
        return self._channels

    def __str__(self):
        base_str = super().__str__()
        with self._state_lock:
            if not self.online:
                return base_str
            doors = "Doors -> "
            doors += ",".join(["%d = %s" % (k, "OPEN" if v else "CLOSED") for k, v in enumerate(self.get_status())])
            return base_str + doors



from enum import Enum


class MerossEventType(Enum):
    # Fired when the MQTT client connects/disconnects to the MQTT broker
    CLIENT_CONNECTION = 10
    DEVICE_ONLINE_STATUS = 100
    DEVICE_SWITCH_STATUS = 1000
    GARAGE_DOOR_STATUS = 2000


class MerossEvent(object):
    event_type = None  # type: MerossEventType

    def __init__(self, event_type):
        self.event_type = event_type


class ClientConnectionEvent(MerossEvent):
    status = None

    def __init__(self, current_status):
        super(ClientConnectionEvent, self).__init__(MerossEventType.CLIENT_CONNECTION)
        self.status = current_status


class DeviceOnlineStatusEvent(MerossEvent):
    # Pointer to the device object
    device = None

    # Current status of the device
    status = None

    def __init__(self, dev, current_status):
        super(DeviceOnlineStatusEvent, self).__init__(MerossEventType.DEVICE_ONLINE_STATUS)
        self.device = dev
        self.status = "online" if current_status else "offline"


class DeviceSwitchStatusEvent(MerossEvent):
    # Pointer to the device object
    device = None

    # Channel ID where the event occurred
    channel_id = None

    # Current state of the switch where the event occurred
    switch_state = None

    # Indicates id the event was generated by a command issued by the library itself.
    # This is particularly useful in the case the user handler wants only to react
    # to events generated by third parties.
    generated_by_myself = None

    def __init__(self, dev, channel_id, switch_state, generated_by_myself):
        super(DeviceSwitchStatusEvent, self).__init__(MerossEventType.DEVICE_SWITCH_STATUS)
        self.device = dev
        self.channel_id = channel_id
        self.generated_by_myself = generated_by_myself


class DeviceDoorStatusEvent(MerossEvent):
    # Pointer to the device object
    device = None

    # Current state of the door
    door_state = None

    # Channel related to the door controller
    channel = None

    # Indicates id the event was generated by a command issued by the library itself.
    # This is particularly useful in the case the user handler wants only to react
    # to events generated by third parties.
    generated_by_myself = None

    def __init__(self, dev, channel_id, door_state, generated_by_myself):
        super(DeviceDoorStatusEvent, self).__init__(MerossEventType.GARAGE_DOOR_STATUS)
        self.device = dev
        self.channel = channel_id
        self.door_state = "open" if door_state else "closed"
        self.generated_by_myself = generated_by_myself



from meross_iot.cloud.abilities import *
from meross_iot.cloud.device import AbstractMerossDevice
from meross_iot.logger import BULBS_LOGGER as l


def to_rgb(rgb):
    if rgb is None:
        return None
    elif isinstance(rgb, int):
        return rgb
    elif isinstance(rgb, tuple):
        red, green, blue = rgb
    elif isinstance(rgb, dict):
        red = rgb['red']
        green = rgb['green']
        blue = rgb['blue']
    else:
        raise Exception("Invalid value for RGB!")

    r = red << 16
    g = green << 8
    b = blue

    return r+g+b


class GenericBulb(AbstractMerossDevice):
    # Bulb state: dictionary of channel-id/bulb-state
    _state = None

    def __init__(self, cloud_client, device_uuid, **kwords):
        self._state = {}
        super(GenericBulb, self).__init__(cloud_client, device_uuid, **kwords)

    def _channel_control_impl(self, channel, status):
        if TOGGLE in self.get_abilities():
            return self._toggle(status)
        elif TOGGLEX in self.get_abilities():
            return self._togglex(channel, status)
        else:
            raise Exception("The current device does not support neither TOGGLE nor TOGGLEX.")

    def _update_state(self, channel, **kwargs):
        with self._state_lock:
            if not channel in self._state:
                self._state[channel] = {}
            for k in kwargs:
                if k == 'onoff':
                    self._state[channel]['onoff'] = kwargs[k]
                elif kwargs[k] is not None:
                    self._state[channel][k] = kwargs[k]

    def _toggle(self, status):
        payload = {"channel": 0, "toggle": {"onoff": status}}
        return self.execute_command("SET", TOGGLE, payload)

    def _togglex(self, channel, status):
        payload = {'togglex': {"onoff": status, "channel": channel}}
        return self.execute_command("SET", TOGGLEX, payload)

    def _handle_push_notification(self, namespace, payload, from_myself=False):
        if namespace == TOGGLE:
            on_status=payload['toggle']['onoff'] == 1
            self._update_state(channel=0, on=on_status)

        elif namespace == TOGGLEX:
            if isinstance(payload['togglex'], list):
                for c in payload['togglex']:
                    channel_index = c['channel']
                    on_status = c['onoff']==1
                    self._update_state(channel=channel_index, on=on_status)
            elif isinstance(payload['togglex'], dict):
                channel_index = payload['togglex']['channel']
                on_status = payload['togglex']['onoff']==1
                self._update_state(channel=channel_index, on=on_status)

        elif namespace == LIGHT:
            c = payload['light']['channel']
            self._update_state(channel=c, kwargs=payload['light'])

        elif namespace == REPORT:
            # For now, we simply ignore push notification of these kind.
            # In the future, we might think of handling such notification by caching them
            # and avoid the network round-trip when asking for power consumption (if the latest report is
            # recent enough)
            pass

        else:
            l.error("Unknown/Unsupported namespace/command: %s" % namespace)

    def _get_status_impl(self):
        res = {}
        data = self.get_sys_data()['all']
        if 'digest' in data:
            light_channel = data['digest']['light']['channel']
            res[light_channel] = data['digest']['light']

            for c in data['digest']['togglex']:
                res[c['channel']]['on'] = c['onoff'] == 1
        elif 'control' in data:
            res[0]['on'] = data['control']['toggle']['onoff'] == 1
        return res

    def _get_channel_id(self, channel):
        # Otherwise, if the passed channel looks like the channel spec, lookup its array indexindex
        if channel in self._channels:
            return self._channels.index(channel)

        # if a channel name is given, lookup the channel id from the name
        if isinstance(channel, str):
            for i, c in enumerate(self.get_channels()):
                if c['devName'] == channel:
                    return c['channel']

        # If an integer is given assume that is the channel ID
        elif isinstance(channel, int):
            return channel

        # In other cases return an error
        raise Exception("Invalid channel specified.")

    def get_status(self, channel=0):
        # In order to optimize the network traffic, we don't call the get_status() api at every request.
        # On the contrary, we call it the first time. Then, the rest of the API will silently listen
        # for state changes and will automatically update the self._state structure listening for
        # messages of the device.
        # Such approach, however, has a side effect. If we call TOGGLE/TOGGLEX and immediately after we call
        # get_status(), the reported status will be still the old one. This is a race condition because the
        # "status" RESPONSE will be delivered some time after the TOGGLE REQUEST. It's not a big issue for now,
        # and synchronizing the two things would be inefficient and probably not very useful.
        # Just remember to wait some time before testing the status of the item after a toggle.
        c = self._get_channel_id(channel)
        if self._state == {}:
            current_state = self._get_status_impl()
            with self._state_lock:
                self._state = current_state
        return self._state[c]

    def get_channels(self):
        return self._channels

    def get_channel_status(self, channel):
        ch_id = self._get_channel_id(channel)
        c = self._get_channel_id(ch_id)
        return self.get_status(c)

    def turn_on_channel(self, channel):
        ch_id = self._get_channel_id(channel)
        c = self._get_channel_id(ch_id)
        return self._channel_control_impl(c, 1)

    def turn_off_channel(self, channel):
        ch_id = self._get_channel_id(channel)
        c = self._get_channel_id(ch_id)
        return self._channel_control_impl(c, 0)

    def turn_on(self, channel=0):
        ch_id = self._get_channel_id(channel)
        c = self._get_channel_id(ch_id)
        return self._channel_control_impl(c, 1)

    def turn_off(self, channel=0):
        ch_id = self._get_channel_id(channel)
        c = self._get_channel_id(ch_id)
        return self._channel_control_impl(c, 0)

    def set_light_color(self, channel=0, rgb=None, luminance=100, temperature=100, capacity=5):
        ch_id = self._get_channel_id(channel)

        # Convert the RGB to integer
        color = to_rgb(rgb)

        payload = {
            'light': {
                'capacity': capacity,
                'channel': ch_id,
                'gradual': 0,
                'luminance': luminance,
                'rgb': color,
                'temperature': temperature
            }
        }

        # TODO: fix this as soon as we get hands on a real MSS560 and see what the payload looks like...
        # handle mss560m differently
        if self.type.lower() == 'mss560m':
            pl = {
                'light': self.get_light_color()
            }
            pl['light']['channel'] = channel
            pl['light']['luminance'] = luminance
            payload = pl

        self.execute_command(command='SET', namespace=LIGHT, payload=payload)

    def get_light_color(self, channel=0):
        ch_id = self._get_channel_id(channel)
        return self.get_status(channel=ch_id)

    def __str__(self):
        base_str = super().__str__()
        with self._state_lock:
            if not self.online:
                return base_str
            channels = "Channels: "
            channels += ",".join(["%d = %s" % (k, "ON" if v else "OFF") for k, v in enumerate(self._state)])
            return base_str + "\n" + "\n" + channels
